'use strict';

var index = require('./options/index.cjs');
var css = require('@master/css');
var extractLatentClasses = require('./functions/extract-latent-classes.cjs');
var fs = require('fs');
var minimatch = require('minimatch');
var log = require('@techor/log');
var extend = require('@techor/extend');
var exploreConfig = require('explore-config');
var exploreCSSConfig = require('@master/css-explore-config');
var cssValidator = require('@master/css-validator');
var chokidar = require('chokidar');
var node_events = require('node:events');
var cssEscape = require('./shared/utils/css-escape.cjs');
var glob = require('@techor/glob');
var path = require('path');

// eslint-disable-next-line @typescript-eslint/no-unsafe-declaration-merging
class CSSExtractor extends node_events.EventEmitter {
    customOptions;
    cwd;
    latentClasses;
    validClasses;
    invalidClasses;
    watching;
    watchers;
    initialized;
    constructor(customOptions = 'master.css-extractor', cwd = process.cwd()){
        super();
        this.customOptions = customOptions;
        this.cwd = cwd;
        this.latentClasses = new Set();
        this.validClasses = new Set();
        this.invalidClasses = new Set();
        this.watching = false;
        this.watchers = [];
        this.initialized = false;
    }
    init(customOptions = this.customOptions) {
        if (this.initialized) return;
        if (typeof customOptions === 'string') {
            this.options = extend(index, exploreConfig(customOptions, {
                found: (basename)=>log.i`Loaded **${basename}**`,
                cwd: this.cwd
            }), customOptions);
        } else {
            this.options = extend(index, customOptions);
        }
        if (this.options.verbose && this.options.verbose > 1) {
            log.ok`**options**`;
            log.tree(this.options);
            log``;
        }
        this.css = new css.MasterCSS(typeof this.options.config === 'object' ? this.options.config : exploreCSSConfig({
            name: this.options.config,
            cwd: this.cwd
        }));
        this.emit('init', this.options, this.config);
        this.initialized = true;
        return this;
    }
    async reset(customOptions = this.customOptions) {
        if (this.watching) await this.disableWatch();
        this.latentClasses.clear();
        this.validClasses.clear();
        this.invalidClasses.clear();
        this.initialized = false;
        this.init(customOptions);
        await this.prepare();
        if (this.watching) await this.initWatch();
        this.emit('reset');
        return this;
    }
    async destroy() {
        this.latentClasses.clear();
        this.validClasses.clear();
        this.invalidClasses.clear();
        this.removeAllListeners();
        await this.closeWatch();
        this.emit('destroy');
        return this;
    }
    async prepare() {
        /* 插入指定的固定 class */ if (this.options.classes?.fixed?.length) {
            for (const eachFixedClass of this.options.classes.fixed){
                this.css.add(eachFixedClass);
            }
            if (this.options.verbose) {
                log.ok`${this.options.classes.fixed.length} fixed classes inserted ${this.options.classes.fixed}`;
            }
        }
        await this.insertFiles(this.fixedSourcePaths);
    }
    /**
     * @description Filter based on relative file paths and extract content
     * @param source
     * @param content
     * @returns string[] Latent classes
     */ extract(source, content) {
        if (!source || !content || !this.isSourceAllowed(source)) {
            return [];
        }
        const latentClasses = [];
        for (const eachLatentClasses of extractLatentClasses(content)){
            if (this.latentClasses.has(eachLatentClasses)) {
                continue;
            } else {
                this.latentClasses.add(eachLatentClasses);
                latentClasses.push(eachLatentClasses);
            }
        }
        return latentClasses;
    }
    /**
     * @description Filter based on relative file paths, extract content, and insert
     * @param source
     * @param content
     * @returns string[] Latent classes
     */ async insert(source, content) {
        if (!content) {
            return false;
        }
        let latentClasses = this.extract(source, content);
        if (!latentClasses.length) {
            return false;
        }
        /**
         * 排除已驗證為 invalid 的 extraction
         */ if (this.invalidClasses.size) {
            latentClasses = latentClasses.filter((eachLatentClass)=>!this.invalidClasses.has(eachLatentClass));
        }
        /**
         * 排除已驗證為 valid 的 extraction
         */ if (this.validClasses.size) {
            latentClasses = latentClasses.filter((eachLatentClass)=>!this.validClasses.has(eachLatentClass));
        }
        /* 排除指定的 class */ if (this.options.classes?.ignored?.length) latentClasses = latentClasses.filter((eachLatentClass)=>{
            if (this.options.classes?.ignored) for (const eachIgnoreClass of this.options.classes.ignored){
                if (typeof eachIgnoreClass === 'string') {
                    if (eachIgnoreClass === eachLatentClass) return false;
                } else if (eachIgnoreClass.test(eachLatentClass)) {
                    return false;
                }
            }
            return true;
        });
        let time = process.hrtime();
        /* 根據類名尋找並插入規則 ( MasterCSS 本身帶有快取機制，重複的類名不會再編譯及產生 ) */ const validClasses = [];
        await Promise.all(latentClasses.map(async (eachLatentClass)=>{
            const validRules = cssValidator.generateValidRules(eachLatentClass, this.css);
            if (validRules.length) {
                for (const validRule of validRules){
                    this.css.insert(validRule);
                }
                validClasses.push(eachLatentClass);
                this.validClasses.add(eachLatentClass);
            } else {
                this.invalidClasses.add(eachLatentClass);
            }
        }));
        time = process.hrtime(time);
        const spent = Math.round((time[0] * 1e9 + time[1]) / 1e6 * 10) / 10;
        if (this.css.rules.length && validClasses.length) {
            if (this.options.verbose) {
                log.ok`**${path.relative(this.cwd, source)}** ${validClasses.length} classes inserted ${log.chalk.gray('in')} ${spent}ms ${this.options.verbose > 1 ? validClasses : ''}`;
            }
            this.emit('change');
        }
        return true;
    }
    insertFile(source) {
        return this.insert(source, fs.readFileSync(path.resolve(this.cwd, source), {
            encoding: 'utf-8'
        }).toString());
    }
    insertFiles(sources) {
        return Promise.all(sources.map((eachRelPaths)=>this.insertFile(eachRelPaths)));
    }
    export(filename = this.options.module) {
        const filepath = path.resolve(this.cwd, filename);
        const dir = path.dirname(filepath);
        if (!fs.existsSync(dir)) {
            fs.mkdirSync(dir, {
                recursive: true
            });
        }
        fs.writeFileSync(filepath, this.css.text);
        if (this.options.verbose) {
            log.success`${this.css.rules.length} rules exported ${log.chalk.gray('in')} **${filename}**`;
        }
        this.emit('export', filename, filepath);
    }
    async watchSource(paths, watchOptions) {
        await this.watch('add change', paths, (source)=>this.insertFile(source), watchOptions);
    }
    async watch(events, paths, handle, watchOptions) {
        watchOptions = extend({
            ignoreInitial: true,
            cwd: this.cwd
        }, watchOptions);
        const watcher = chokidar.watch(paths, watchOptions);
        this.watchers.push(watcher);
        events.split(' ').forEach((eachEvent)=>watcher.on(eachEvent, handle));
        await new Promise((resolve)=>{
            watcher.once('ready', resolve);
        });
    }
    async initWatch() {
        const resolvedConfigPath = this.resolvedConfigPath;
        const resolvedOptionsPath = this.resolvedOptionsPath;
        if (this.options.sources?.length) {
            await this.watchSource(this.options.sources);
        }
        if (resolvedConfigPath) {
            await this.watch('add change unlink', resolvedConfigPath, async ()=>{
                if (this.options.verbose) {
                    log``;
                    log`[change] **${this.configPath}**`;
                }
                await this.reset();
                this.emit('configChange');
            });
        }
        if (resolvedOptionsPath) {
            await this.watch('add change unlink', resolvedOptionsPath, async ()=>{
                if (this.options.verbose) {
                    log``;
                    log`[change] **${this.customOptions}**`;
                }
                await this.reset();
                this.emit('optionsChange');
            });
        }
    }
    async disableWatch() {
        if (this.watchers.length) {
            await Promise.all(this.watchers.map((eachWatcher)=>eachWatcher.removeAllListeners()));
            this.watchers.length = 0;
        }
    }
    async startWatch() {
        if (this.watching) return;
        await this.initWatch();
        this.watching = true;
        this.emit('watchStart');
    }
    async closeWatch() {
        if (!this.watching) return;
        if (this.watchers.length) {
            await Promise.all(this.watchers.map((eachWatcher)=>eachWatcher.close()));
            this.watchers.length = 0;
        }
        this.watching = false;
        this.emit('watchClose');
    }
    /**
     * computed from `options.sources`
     */ get fixedSourcePaths() {
        const { sources } = this.options;
        return sources?.length ? glob.explorePathsSync(sources, {
            cwd: this.cwd
        }).filter((eachSourcePath)=>!!eachSourcePath) : [];
    }
    /**
     * resolved from `fixedSourcePaths`
     */ get resolvedFixedSourcePaths() {
        return this.fixedSourcePaths.map((eachSourcePath)=>path.resolve(this.cwd, eachSourcePath));
    }
    /**
     * `options.include` - `options.exclude`
     */ get allowedSourcePaths() {
        const { include, exclude } = this.options;
        return include?.length ? glob.explorePathsSync(include, {
            cwd: this.cwd,
            ignore: exclude
        }).filter((eachSourcePath)=>!!eachSourcePath) : [];
    }
    /**
     * resolved from `allowedSourcePaths`
     */ get resolvedAllowedSourcePaths() {
        return this.allowedSourcePaths.map((eachSourcePath)=>path.resolve(this.cwd, eachSourcePath));
    }
    isSourceAllowed(source) {
        /* remove if params exists */ if (source.includes('?')) {
            source = source.split('?')[0];
        }
        const { include, exclude, sources } = this.options;
        if (sources) for (const eachSource of sources){
            if (minimatch.minimatch(source, eachSource, {
                dot: true
            })) return true;
        }
        if (include) for (const eachIncludePattern of include){
            if (!minimatch.minimatch(source, eachIncludePattern, {
                dot: true
            })) return false;
        }
        if (exclude) for (const eachExcludePattern of exclude){
            if (minimatch.minimatch(source, eachExcludePattern, {
                dot: true
            })) return false;
        }
        return true;
    }
    /**
     * computed from `options.config`
     */ get config() {
        return this.css.config;
    }
    /**
     * computed from string `options.config`
    */ get configPath() {
        if (typeof this.options.config === 'string') {
            let foundBasename;
            for (const eachExtension of [
                'js',
                'mjs',
                'ts',
                'cjs',
                'cts',
                'mts'
            ]){
                const eachBasename = this.options.config + '.' + eachExtension;
                const eachPath = path.resolve(this.cwd || '', eachBasename);
                if (fs.existsSync(eachPath)) {
                    foundBasename = eachBasename;
                    break;
                }
            }
            return foundBasename;
        }
    }
    /**
     * computed from string `options.config`
    */ get resolvedConfigPath() {
        const configPath = this.configPath;
        if (configPath) {
            return path.resolve(this.cwd, configPath);
        }
    }
    /**
     * computed from string `customOptions`
    */ get optionsPath() {
        if (typeof this.customOptions === 'string') {
            let foundBasename;
            for (const eachExtension of [
                'js',
                'mjs',
                'ts',
                'cjs',
                'cts',
                'mts'
            ]){
                const eachBasename = this.customOptions + '.' + eachExtension;
                const eachPath = path.resolve(this.cwd || '', eachBasename);
                if (fs.existsSync(eachPath)) {
                    foundBasename = eachBasename;
                    break;
                }
            }
            return foundBasename;
        }
    }
    /**
     * computed from string `customOptions`
    */ get resolvedOptionsPath() {
        const optionsPath = this.optionsPath;
        if (optionsPath) {
            return path.resolve(this.cwd, optionsPath);
        }
    }
    get resolvedVirtualModuleId() {
        return '\0' + this.options.module;
    }
    get slotCSSRule() {
        return '#' + cssEscape(this.options.module) + '{--slot:0}';
    }
}

module.exports = CSSExtractor;
